/* 
* Formal verification of the quick sort algorithm with Dafny.
*/

// Checks if array 'a' is sorted between positions 'lo' and 'hi' (inclusive)
predicate sorted(a: array<int>, lo: int, hi: int)
reads a
{
    forall i, j :: 0 <= lo <= i < j <= hi < a.Length ==> a[i] <= a[j] 
}

// Checks if values of array 'a' in the range [0 .. lo-1] <= [lo..hi] <= [hi+1..a.Length-1] 
predicate partitioned(a: array<int>, lo: int, hi: int)
reads a
{
    (forall i, j :: 0 <= i < lo <= j <= hi < a.Length ==> a[i] <= a[j]) 
    && (forall i, j :: 0 <= lo <= i <= hi < j < a.Length ==> a[i] <= a[j]) 
}

// Sorts array 'a' using the quicksort algorithm
method quicksort(a: array<int>)
  modifies a
  ensures sorted(a, 0, a.Length-1)
  ensures multiset(a[..]) == multiset(old(a)[..])
{
    quicksort2(a, 0, a.Length-1);
}

// Recursive method for performing quicksort of array 'a' between indices 'lo' and 'hi' (inclusive).
method quicksort2(a: array<int>, lo: int, hi: int)
  requires 0 <= lo <= hi + 1 <= a.Length
  modifies a
 decreases hi - lo
  
{
    if lo < hi
    {
        var pivot := partition(a, lo, hi);
        quicksort2(a, lo, pivot  - 1);
        quicksort2(a, pivot + 1, hi);
    }
}

// Partitions a non-empty subarray 'a' between indices 'lo' and 'hi', with smaller values
// to the left and higher values to the right of the pivot, and returns the pivot position.
method partition(a: array<int>, lo: int, hi: int) returns(pivot: int)
  requires 0 <= lo <= hi < a.Length
  requires partitioned(a, lo, hi)
  modifies a
{
    var i := lo;
    var j := lo;
    pivot := hi;    
    while  j < hi
      decreases hi - j

    {
        if a[j] < a[hi]
        {
            a[i], a[j] := a[j], a[i];
            i := i + 1;
        }
        j := j+1;
    }
    a[hi], a[i] := a[i], a[hi];
    pivot := i;
    return pivot;
}

method printArray<T>(a: array<T>)
{
    var i := 0;
    print "{";
    while (i < a.Length)
     decreases a.Length - i;
    {
        if (i > 0)
        {
            print ", ";
        }
        print a[i];
        i := i+1;
    }
    print "}\n" ;
}

method Main() {
  var a := new int[5];
  a[0], a[1], a[2], a[3], a[4] := 9, 4, 6, 4, -1;
  print "antes de ordenar: ";
  printArray(a);  
  quicksort(a);
  print "depois de ordenar: ";
  printArray(a); 
} 
